#include <stdio.h>
#include "hex.h"
#include "hsv.h"
#include "engine.h"
#include "vulkan/vulkan_core.h"

#include <math.h>
#include <stdlib.h>

#define COLOR_PICK_CONTAINER_ID 0x03
typedef struct EditorDataStruct { 
  uint32_t selected_region;

  uint32_t last_clicked_region;
  uint32_t last_clicked_hex;
  uint32_t last_clicked_vertex;

  double current_hsv[3];
  vec4 current;
  vec4 saved[12];
  char string[10];
  int string_len;
} EditorData;


uint32_t hex_color(vec4 color) {
  return ((lrint(color[0]*255.0) & 0xFF) << 24)
       + ((lrint(color[1]*255.0) & 0xFF) << 16)
       + ((lrint(color[2]*255.0) & 0xFF) <<  8)
       + ((lrint(color[3]*255.0) & 0xFF) <<  0);
}

VkResult set_vertex_height(float height, uint32_t region, uint32_t hex, uint32_t vertex, ClientContext* context) {
  if(vertex == 0) return VK_ERROR_VALIDATION_FAILED_EXT;
  context->hex.regions[region]->data.hexes[hex].height[vertex-1] = height; 
  return add_transfer(
      &context->hex.regions[region]->data.hexes[hex].height[vertex-1],
      context->hex.regions[region]->region,
      offsetof(GPUHexRegion, hexes)
      + sizeof(GPUHex)*hex
      + offsetof(GPUHex, height)
      + sizeof(float)*(vertex-1),
      sizeof(float),
      context->render.current_frame,
      &context->render);
}

VkResult set_vertex_color(vec4 color, uint32_t region, uint32_t hex, uint32_t vertex, ClientContext* context) {
  context->hex.regions[region]->data.hexes[hex].color[vertex] = hex_color(color); 
  return add_transfer(
      &context->hex.regions[region]->data.hexes[hex].color[vertex],
      context->hex.regions[region]->region,
      offsetof(GPUHexRegion, hexes)
      + sizeof(GPUHex)*hex
      + offsetof(GPUHex, color)
      + sizeof(uint32_t)*vertex,
      sizeof(uint32_t),
      context->render.current_frame,
      &context->render);
}

uint32_t add_hex_region(ClientContext* context) {
  HexRegion* region;

  allocate_hex_region(0, 0, 0, context->hex.data.current_map, &region, &context->hex, &context->render);

  for(uint32_t hex = 0; hex < REGION_HEX_COUNT; hex++) {
    region->data.hexes[hex].color[0] = 0xFFFFFFFF;
    region->data.hexes[hex].color[1] = 0xFFFFFFFF;
    region->data.hexes[hex].color[2] = 0xFFFFFFFF;
    region->data.hexes[hex].color[3] = 0xFFFFFFFF;
    region->data.hexes[hex].color[4] = 0xFFFFFFFF;
    region->data.hexes[hex].color[5] = 0xFFFFFFFF;
    region->data.hexes[hex].color[6] = 0xFFFFFFFF;

    region->data.hexes[hex].height[0] = 0;
    region->data.hexes[hex].height[1] = 0;
    region->data.hexes[hex].height[2] = 0;
    region->data.hexes[hex].height[3] = 0;
    region->data.hexes[hex].height[4] = 0;
    region->data.hexes[hex].height[5] = 0;
  }

  set_hex_region(region, &context->hex, &context->render);

  uint32_t i = 0;
  for(; i < MAX_LOADED_REGIONS; i++) {
    if(region == context->hex.regions[i]) {
      break;
    }
  }

  return i;
}

void update_hex_string(UIContext* ui, RenderContext* gpu, EditorData* data) {
  snprintf(data->string, 10, "#%02X%02X%02X%02X",
      (uint)rintf(data->current[0]*255),
      (uint)rintf(data->current[1]*255),
      (uint)rintf(data->current[2]*255),
      (uint)rintf(data->current[3]*255));
  data->string_len = 8;
  update_ui_string(data->string, COLOR_PICK_CONTAINER_ID, 0, 0, ui, gpu);
}

void sv_square_pick(UIContext* ui, RenderContext* gpu, EditorData* data, float s, float v) {
  if(s < 0) s = 0;
  if(s > 1) s = 1;

  if(v < 0) v = 0;
  if(v > 1) v = 1;

  data->current_hsv[1] = s;
  data->current_hsv[2] = v;

  Container* container = context_container(COLOR_PICK_CONTAINER_ID, ui);
  Layer* layer = &container->layers[0];
  GPUDrawable* select_outline = &layer->drawables_buffer[3];
  GPUDrawable* select = &layer->drawables_buffer[4];

  select->pos[0] = s*130 - 2;
  select->pos[1] = 130 - v*130 - 2;
  
  select->color[0][1] = s;
  select->color[0][2] = v;

  select->color[1][1] = s;
  select->color[1][2] = v;

  select->color[2][1] = s;
  select->color[2][2] = v;

  select->color[3][1] = s;
  select->color[3][2] = v;

  select_outline->pos[0] = s*130 - 3;
  select_outline->pos[1] = 130 - v*130 - 3;

  add_transfers(
      &select_outline->pos[0],
      layer->drawables,
      3*sizeof(GPUDrawable) + offsetof(GPUDrawable, pos),
      1*sizeof(vec2),
      gpu);

  add_transfers(
      &select->pos[0],
      layer->drawables,
      4*sizeof(GPUDrawable) + offsetof(GPUDrawable, pos),
      1*sizeof(vec2),
      gpu);

  add_transfers(
      &select->color[0],
      layer->drawables,
      4*sizeof(GPUDrawable) + offsetof(GPUDrawable, color),
      4*sizeof(vec4),
      gpu);

  hsv_to_rgb(data->current_hsv, data->current);
  update_hex_string(ui, gpu, data);
}

bool sv_square_button_callback(void* data, UIContext* ui, RenderContext* gpu, float x, float y, int button, int action, int mods) {
  (void)mods;
  (void)x;

  if(action == GLFW_PRESS && button == GLFW_MOUSE_BUTTON_LEFT) {
    set_active_element(COLOR_PICK_CONTAINER_ID, 0, 1, ui);
    sv_square_pick(ui, gpu, data, x, 1-y);
  } else if(action == GLFW_RELEASE && button == GLFW_MOUSE_BUTTON_LEFT) {
    clear_active_element(ui, gpu);
  }

  return true;
}

bool sv_square_cursor_callback(void* data, UIContext* ui, RenderContext* gpu, float x, float y) {
  if(ui->active_element == 1
  && ui->active_layer == 0
  && ui->active_container == COLOR_PICK_CONTAINER_ID) {
    sv_square_pick(ui, gpu, data, x, 1-y);
  }

  return true;
}

void hue_bar_set(UIContext* ui, RenderContext* gpu, EditorData* data, float y) {
  if(y < 0) y = 0;
  if(y > 1) y = 1;

  data->current_hsv[0] = y;

  Container* container = context_container(COLOR_PICK_CONTAINER_ID, ui);
  Layer* layer = &container->layers[0];
  GPUDrawable* sv_square = &layer->drawables_buffer[1];
  GPUDrawable* sv_select = &layer->drawables_buffer[4];
  GPUDrawable* hue_select = &layer->drawables_buffer[6];
  
  sv_square->color[0][0] = y;
  sv_square->color[1][0] = y;
  sv_square->color[2][0] = y;
  sv_square->color[3][0] = y;

  sv_select->color[0][0] = y;
  sv_select->color[1][0] = y;
  sv_select->color[2][0] = y;
  sv_select->color[3][0] = y;

  hue_select->pos[1] = 2 + y*130;

  add_transfers(
      &sv_square->color[0],
      layer->drawables,
      1*sizeof(GPUDrawable) + offsetof(GPUDrawable, color),
      4*sizeof(vec4),
      gpu);

  add_transfers(
      &sv_select->color[0],
      layer->drawables,
      4*sizeof(GPUDrawable) + offsetof(GPUDrawable, color),
      4*sizeof(vec4),
      gpu);

  add_transfers(
      &hue_select->pos[1],
      layer->drawables,
      6*sizeof(GPUDrawable) + offsetof(GPUDrawable, pos) + sizeof(float),
      1*sizeof(float),
      gpu);

  hsv_to_rgb(data->current_hsv, data->current);
  update_hex_string(ui, gpu, data);
}

bool hue_bar_scroll_callback(void* data, UIContext* ui, RenderContext* gpu, double x, double y) {
  (void)x;

  Container* container = context_container(COLOR_PICK_CONTAINER_ID, ui);
  hue_bar_set(ui, gpu, data, y*0.01 + container->layers[0].drawables_buffer[1].color[0][0]);

  return true;
}

bool hue_bar_cursor_callback(void*data, UIContext* ui, RenderContext* gpu, float x, float y) {
  (void)x;

  if(ui->active_element == 2
  && ui->active_layer == 0
  && ui->active_container == COLOR_PICK_CONTAINER_ID) {
    hue_bar_set(ui, gpu, data, y);
  }

  return true;
}

bool hue_bar_button_callback(void* data, UIContext* ui, RenderContext* gpu, float x, float y, int button, int action, int mods) {
  (void)mods;
  (void)x;

  if(action == GLFW_PRESS && button == GLFW_MOUSE_BUTTON_LEFT) {
    set_active_element(COLOR_PICK_CONTAINER_ID, 0, 2, ui);
    hue_bar_set(ui, gpu, data, y);
  } else if(action == GLFW_RELEASE && button == GLFW_MOUSE_BUTTON_LEFT) {
    clear_active_element(ui, gpu);
  }

  return true;
}

void hex_string_set_color(UIContext* ui, RenderContext* gpu, float color) {
  Container* container = context_container(COLOR_PICK_CONTAINER_ID, ui);
  container->layers[0].drawables_buffer[5].color[0][2] = color;
  container->layers[0].drawables_buffer[5].color[0][3] = color;

  container->layers[0].drawables_buffer[5].color[1][2] = color;
  container->layers[0].drawables_buffer[5].color[1][3] = color;

  container->layers[0].drawables_buffer[5].color[2][2] = color;
  container->layers[0].drawables_buffer[5].color[2][3] = color;

  container->layers[0].drawables_buffer[5].color[3][2] = color;
  container->layers[0].drawables_buffer[5].color[3][3] = color;

  add_transfers(
      &container->layers[0].drawables_buffer[5].color[0],
      container->layers[0].drawables,
      5*sizeof(GPUDrawable) + offsetof(GPUDrawable, color),
      4*sizeof(vec4),
      gpu);
}

bool hex_string_text_callback(void* ptr, UIContext* ui, RenderContext* gpu, unsigned int codepoint) {
  EditorData* data = ptr;

  if(codepoint >= 'a' && codepoint <= 'f') {
    codepoint += 'A' - 'a';
  } else if(!((codepoint >= 'A' && codepoint <= 'F') || (codepoint >= '0' && codepoint <= '9'))) {
    return true;
  }

  if(data->string_len < 8) {
    data->string_len += 1;
    data->string[data->string_len] = codepoint;
    update_ui_string(data->string, COLOR_PICK_CONTAINER_ID, 0, 0, ui, gpu);
  }

  return true;
}

bool hex_string_key_callback(void* ptr, UIContext* ui, RenderContext* gpu, int key, int action, int mods) {
  (void)mods;
  EditorData* data = ptr;
  char tmp[3];

  if(action == GLFW_PRESS) {
    switch(key) {
    case GLFW_KEY_ESCAPE:
      hsv_to_rgb(data->current_hsv, data->current);
      update_hex_string(ui, gpu, data);
      clear_active_element(ui, gpu);
      break;
    case GLFW_KEY_ENTER:
      // TODO: validate hex string and reset or set hsv
      tmp[0] = data->string[1];
      tmp[1] = data->string[2];
      tmp[2] = 0;
      data->current[0] = strtol(tmp, NULL, 16)/255.0;

      tmp[0] = data->string[3];
      tmp[1] = data->string[4];
      tmp[2] = 0;
      data->current[1] = strtol(tmp, NULL, 16)/255.0;

      tmp[0] = data->string[5];
      tmp[1] = data->string[6];
      tmp[2] = 0;
      data->current[2] = strtol(tmp, NULL, 16)/255.0;

      tmp[0] = data->string[7];
      tmp[1] = data->string[8];
      tmp[2] = 0;
      data->current[3] = strtol(tmp, NULL, 16)/255.0;

      rgb_to_hsv(data->current, data->current_hsv);
      hue_bar_set(ui, gpu, data, data->current_hsv[0]);
      sv_square_pick(ui, gpu, data, data->current_hsv[1], data->current_hsv[2]);

      clear_active_element(ui, gpu);
      break;
    case GLFW_KEY_BACKSPACE:
      if(data->string_len > 0) {
        data->string[data->string_len] = '\0';
        data->string_len -= 1;
        update_ui_string(data->string, COLOR_PICK_CONTAINER_ID, 0, 0, ui, gpu);
      }
      break;
    }
  }

  return true;
}

bool hex_string_button_callback(void* data, UIContext* ui, RenderContext* gpu, float x, float y, int button, int action, int mods) {
  (void)data;
  (void)mods;
  (void)x;
  (void)y;

  if(action == GLFW_PRESS && button == GLFW_MOUSE_BUTTON_LEFT) {
    set_active_element(COLOR_PICK_CONTAINER_ID, 0, 5, ui);
    hex_string_set_color(ui, gpu, 1);
  }

  return true;
}

void hex_string_deselect_callback(void* data, UIContext* ui, RenderContext* gpu) {
  (void)data;
  hex_string_set_color(ui, gpu, 0);
}

void set_saved_color(UIContext* ui, RenderContext* gpu, EditorData* data, uint32_t index, vec4 color) {
  data->saved[index][0] = color[0];
  data->saved[index][1] = color[1];
  data->saved[index][2] = color[2];
  data->saved[index][3] = color[3];

  Container* container = context_container(COLOR_PICK_CONTAINER_ID, ui);
  Layer* layer = &container->layers[0];
  GPUDrawable* saved_color = &layer->drawables_buffer[7 + index];

  saved_color->color[0][0] = data->current[0];
  saved_color->color[0][1] = data->current[1];
  saved_color->color[0][2] = data->current[2];

  saved_color->color[1][0] = data->current[0];
  saved_color->color[1][1] = data->current[1];
  saved_color->color[1][2] = data->current[2];

  saved_color->color[2][0] = data->current[0];
  saved_color->color[2][1] = data->current[1];
  saved_color->color[2][2] = data->current[2];

  saved_color->color[3][0] = data->current[0];
  saved_color->color[3][1] = data->current[1];
  saved_color->color[3][2] = data->current[2];

  add_transfers(
      &saved_color->color[0],
      layer->drawables,
      (7+index)*sizeof(GPUDrawable) + offsetof(GPUDrawable, color),
      4*sizeof(vec4),
      gpu);
}

bool saved_color_button_callback(
    void* ptr,
    UIContext* ui,
    RenderContext* gpu,
    float x, float y,
    int button, int action, int mods,
    uint32_t index) {
  (void)x;
  (void)y;
  (void)mods;

  EditorData* data = ptr;

  if(action == GLFW_PRESS) {
    if(button == GLFW_MOUSE_BUTTON_LEFT) {
      data->current[0] = data->saved[index][0];
      data->current[1] = data->saved[index][1];
      data->current[2] = data->saved[index][2];
      data->current[3] = data->saved[index][3];
      rgb_to_hsv(data->current, data->current_hsv);
      hue_bar_set(ui, gpu, data, data->current_hsv[0]);
      sv_square_pick(ui, gpu, data, data->current_hsv[1], data->current_hsv[2]);
      update_hex_string(ui, gpu, data);
    } else if(button == GLFW_MOUSE_BUTTON_RIGHT) {
      set_saved_color(ui, gpu, data, index, data->current);
    } else if (button == GLFW_MOUSE_BUTTON_MIDDLE) {
      vec4 clear = {0, 0, 0, 0};
      set_saved_color(ui, gpu, data, index, clear);
    }
  }

  return true;
}

#define SAVED_COLOR_BUTTON_CALLBACK(n) \
bool saved_color_button_callback_##n(void* data, UIContext* ui, RenderContext* gpu, float x, float y, int button, int action, int mods) { \
  return saved_color_button_callback(data, ui, gpu, x, y, button, action, mods, n); \
}

SAVED_COLOR_BUTTON_CALLBACK( 0);
SAVED_COLOR_BUTTON_CALLBACK( 1);
SAVED_COLOR_BUTTON_CALLBACK( 2);
SAVED_COLOR_BUTTON_CALLBACK( 3);
SAVED_COLOR_BUTTON_CALLBACK( 4);
SAVED_COLOR_BUTTON_CALLBACK( 5);
SAVED_COLOR_BUTTON_CALLBACK( 6);
SAVED_COLOR_BUTTON_CALLBACK( 7);
SAVED_COLOR_BUTTON_CALLBACK( 8);
SAVED_COLOR_BUTTON_CALLBACK( 9);
SAVED_COLOR_BUTTON_CALLBACK(10);
SAVED_COLOR_BUTTON_CALLBACK(11);

#define SAVED_COLOR_CALLBACK_ENTRY(n) \
    { \
      .layer = 0, \
      .element = 7 + n, \
      .data = data, \
      .button = saved_color_button_callback_##n, \
    }

VkResult color_ui(ClientContext* context) {
  GPUString strings[] = {
    {
      .pos = {2, 150},
      .color = {1, 1, 1, 1},
      .size = 16,
      .offset = 0,
      .length = 9,
      .font = 0,
    },
  };

  GPUDrawable drawables[] = {
    {
      .pos = {0, 0},
      .size = {190, 150},
      .color = {{0.4, 0.4, 0.4, 0.8}, {0.4, 0.4, 0.4, 0.8}, {0.4, 0.4, 0.4, 0.8}, {0.4, 0.4, 0.4, 0.8}},
    },
    {
      .type = DRAWABLE_TYPE_RECT_HSV,
      .pos = {2, 2},
      .size = {130, 130},
      .color = {{0, 0, 1, 1}, {0, 1, 1, 1}, {0, 0, 0, 1}, {0, 1, 0, 1}},
      .events = UI_EVENT_BUTTON,
    },
    {
      .type = DRAWABLE_TYPE_RECT_HSV,
      .pos = {134, 2},
      .size = {10, 130},
      .color = {{0, 1, 1, 1}, {0, 1, 1, 1}, {1, 1, 1, 1}, {1, 1, 1, 1}},
      .events = UI_EVENT_BUTTON | UI_EVENT_SCROLL,
    },
    {
      .pos = {130-4, 130-4},
      .size = {7, 7},
      .color = {{0, 0, 0, 1}, {0, 0, 0, 1}, {0, 0, 0, 1}, {0, 0, 0, 1}},
    },
    {
      .type = DRAWABLE_TYPE_RECT_HSV,
      .pos = {130-3, 130-3},
      .size = {5, 5},
      .color = {{1, 0, 0, 1}, {1, 0, 0, 1}, {1, 0, 0, 1}, {1, 0, 0, 1}},
    },
    {
      .pos = {20, 134},
      .size = {95, 15},
      .events = UI_EVENT_BUTTON | UI_EVENT_CURSOR,
    },
    {
      .pos = {134, 2},
      .size = {10, 1},
      .color = {{0, 0, 0, 1}, {0, 0, 0, 1}, {0, 0, 0, 1}, {0, 0, 0, 1}},
    },
    {
      .pos = {146, 2},
      .size = {20, 20},
      .color = {{0, 0, 0, 1}, {0, 0, 0, 1}, {0, 0, 0, 1}, {0, 0, 0, 1}},
      .events = UI_EVENT_BUTTON,
    },
    {
      .pos = {168, 2},
      .size = {20, 20},
      .color = {{0, 0, 0, 1}, {0, 0, 0, 1}, {0, 0, 0, 1}, {0, 0, 0, 1}},
      .events = UI_EVENT_BUTTON,
    },
    {
      .pos = {146, 24},
      .size = {20, 20},
      .color = {{0, 0, 0, 1}, {0, 0, 0, 1}, {0, 0, 0, 1}, {0, 0, 0, 1}},
      .events = UI_EVENT_BUTTON,
    },
    {
      .pos = {168, 24},
      .size = {20, 20},
      .color = {{0, 0, 0, 1}, {0, 0, 0, 1}, {0, 0, 0, 1}, {0, 0, 0, 1}},
      .events = UI_EVENT_BUTTON,
    },
    {
      .pos = {146, 46},
      .size = {20, 20},
      .color = {{0, 0, 0, 1}, {0, 0, 0, 1}, {0, 0, 0, 1}, {0, 0, 0, 1}},
      .events = UI_EVENT_BUTTON,
    },
    {
      .pos = {168, 46},
      .size = {20, 20},
      .color = {{0, 0, 0, 1}, {0, 0, 0, 1}, {0, 0, 0, 1}, {0, 0, 0, 1}},
      .events = UI_EVENT_BUTTON,
    },
    {
      .pos = {146, 68},
      .size = {20, 20},
      .color = {{0, 0, 0, 1}, {0, 0, 0, 1}, {0, 0, 0, 1}, {0, 0, 0, 1}},
      .events = UI_EVENT_BUTTON,
    },
    {
      .pos = {168, 68},
      .size = {20, 20},
      .color = {{0, 0, 0, 1}, {0, 0, 0, 1}, {0, 0, 0, 1}, {0, 0, 0, 1}},
      .events = UI_EVENT_BUTTON,
    },
    {
      .pos = {146, 90},
      .size = {20, 20},
      .color = {{0, 0, 0, 1}, {0, 0, 0, 1}, {0, 0, 0, 1}, {0, 0, 0, 1}},
      .events = UI_EVENT_BUTTON,
    },
    {
      .pos = {168, 90},
      .size = {20, 20},
      .color = {{0, 0, 0, 1}, {0, 0, 0, 1}, {0, 0, 0, 1}, {0, 0, 0, 1}},
      .events = UI_EVENT_BUTTON,
    },
    {
      .pos = {146, 112},
      .size = {20, 20},
      .color = {{0, 0, 0, 1}, {0, 0, 0, 1}, {0, 0, 0, 1}, {0, 0, 0, 1}},
      .events = UI_EVENT_BUTTON,
    },
    {
      .pos = {168, 112},
      .size = {20, 20},
      .color = {{0, 0, 0, 1}, {0, 0, 0, 1}, {0, 0, 0, 1}, {0, 0, 0, 1}},
      .events = UI_EVENT_BUTTON,
    },
  };

  uint32_t codes[9];
  VkResult result;
  VK_RESULT(map_string("#00000000", codes, 0, 0, &context->ui));

  LayerInput layer = {
    .strings = strings,
    .num_strings = sizeof(strings)/sizeof(GPUString),
    .codes = codes,
    .num_codes = sizeof(codes)/sizeof(uint32_t),
    .drawables = drawables,
    .num_drawables = sizeof(drawables)/sizeof(GPUDrawable),
  };

  EditorData* data = context->app_data;
  data->string_len = 8;
  data->string[0] = '#';
  data->string[1] = '0';
  data->string[2] = '0';
  data->string[3] = '0';
  data->string[4] = '0';
  data->string[5] = '0';
  data->string[6] = '0';
  data->string[7] = 'F';
  data->string[8] = 'F';
  data->string[9] = '\0';

  data->current[0] = 0;
  data->current[1] = 0;
  data->current[2] = 0;
  data->current[3] = 1;

  for(uint32_t i = 0; i < 12; i++) {
    data->saved[i][0] = 0;
    data->saved[i][1] = 0;
    data->saved[i][2] = 0;
    data->saved[i][3] = 1;
  }


  UICallbacks callbacks[] = {
    {
      .layer = 0,
      .element = 1,
      .data = context->app_data,
      .button = sv_square_button_callback,
      .cursor = sv_square_cursor_callback,
    },
    {
      .layer = 0,
      .element = 2,
      .data = context->app_data,
      .button = hue_bar_button_callback,
      .cursor = hue_bar_cursor_callback,
      .scroll = hue_bar_scroll_callback,
    },
    {
      .layer = 0,
      .element = 5,
      .data = context->app_data,
      .button = hex_string_button_callback,
      .key = hex_string_key_callback,
      .text = hex_string_text_callback,
      .deselect = hex_string_deselect_callback,
    },
    SAVED_COLOR_CALLBACK_ENTRY( 0),
    SAVED_COLOR_CALLBACK_ENTRY( 1),
    SAVED_COLOR_CALLBACK_ENTRY( 2),
    SAVED_COLOR_CALLBACK_ENTRY( 3),
    SAVED_COLOR_CALLBACK_ENTRY( 4),
    SAVED_COLOR_CALLBACK_ENTRY( 5),
    SAVED_COLOR_CALLBACK_ENTRY( 6),
    SAVED_COLOR_CALLBACK_ENTRY( 7),
    SAVED_COLOR_CALLBACK_ENTRY( 8),
    SAVED_COLOR_CALLBACK_ENTRY( 9),
    SAVED_COLOR_CALLBACK_ENTRY(10),
    SAVED_COLOR_CALLBACK_ENTRY(11),
  };

  ContainerInput container = {
    .layers = &layer,
    .layer_count = 1,
    .anchor = ANCHOR_BOTTOM_LEFT,
    .id = COLOR_PICK_CONTAINER_ID,
    .offset = {0, 0},
    .size = {190, 150},
    .callbacks = callbacks,
    .callback_count = sizeof(callbacks)/sizeof(UICallbacks),
  };

  return create_container(&container, &context->render, &context->ui);
}

VkResult hex_info_ui(ClientContext* context) {
  GPUString strings[] = {
    {
      .pos = {0, 33},
      .color = {1, 1, 1, 1},
      .size = 32,
      .offset = 0,
      .length = 0,
      .font = 0,
    },
    {
      .pos = {2, 50},
      .color = {1, 1, 1, 1},
      .size = 16,
      .offset = 20,
      .length = 0,
      .font = 0,
    },
    {
      .pos = {2, 73},
      .color = {1, 1, 1, 1},
      .size = 16,
      .offset = 40,
      .length = 0,
      .font = 0,
    },
    {
      .pos = {2, 93},
      .color = {1, 1, 1, 1},
      .size = 16,
      .offset = 60,
      .length = 0,
      .font = 0,
    },
    {
      .pos = {2, 113},
      .color = {1, 1, 1, 1},
      .size = 16,
      .offset = 80,
      .length = 0,
      .font = 0,
    },
    {
      .pos = {2, 133},
      .color = {1, 1, 1, 1},
      .size = 16,
      .offset = 100,
      .length = 0,
      .font = 0,
    },
    {
      .pos = {2, 153},
      .color = {1, 1, 1, 1},
      .size = 16,
      .offset = 120,
      .length = 0,
      .font = 0,
    },
    {
      .pos = {2, 173},
      .color = {1, 1, 1, 1},
      .size = 16,
      .offset = 140,
      .length = 0,
      .font = 0,
    },

  };

  GPUDrawable drawables[] = {
    {
      .pos = {0, 0},
      .size = {150, 175},
      .color = {{0.4, 0.4, 0.4, 0.8}, {0.4, 0.4, 0.4, 0.8}, {0.4, 0.4, 0.4, 0.8}, {0.4, 0.4, 0.4, 0.8}},
    },
  };

  LayerInput layer = {
    .strings = strings,
    .num_strings = sizeof(strings)/sizeof(GPUString),
    .max_strings = sizeof(strings)/sizeof(GPUString) + 4,
    .max_codes = 200,
    .drawables = drawables,
    .num_drawables = sizeof(drawables)/sizeof(GPUDrawable),
  };

  ContainerInput container = {
    .id = 0x02,
    .size = {150, 175},
    .offset = {0, 0},
    .anchor = ANCHOR_BOTTOM_RIGHT,
    .layers = &layer,
    .layer_count = 1,
  };

  return create_container(&container, &context->render, &context->ui);
}

VkResult region_info_ui(ClientContext* context) {

  GPUString strings[] = {
    {
      .pos = {0, 33},
      .color = {1, 1, 1, 1},
      .size = 32,
      .offset = 0,
      .length = 0,
      .font = 0,
    },
    {
      .pos = {0, 33 + 1*40},
      .color = {1, 1, 1, 1},
      .size = 32,
      .offset = 11,
      .length = 0,
      .font = 0,
    },
    {
      .pos = {0, 33 + 2*40},
      .color = {1, 1, 1, 1},
      .size = 32,
      .offset = 18,
      .length = 0,
      .font = 0,
    },
    {
      .pos = {0, 33 + 3*40},
      .color = {1, 1, 1, 1},
      .size = 32,
      .offset = 25,
      .length = 0,
      .font = 0,
    },
  };

  GPUDrawable drawables[] = {
    {
      .pos = {0, 0},
      .size = {225, 155},
      .color = {{0.4, 0.4, 0.4, 0.8}, {0.4, 0.4, 0.4, 0.8}, {0.4, 0.4, 0.4, 0.8}, {0.4, 0.4, 0.4, 0.8}},
    },
  };

  LayerInput layer = {
    .strings = strings,
    .num_strings = sizeof(strings)/sizeof(GPUString),
    .max_strings = sizeof(strings)/sizeof(GPUString) + 4,
    .max_codes = 100,
    .drawables = drawables,
    .num_drawables = sizeof(drawables)/sizeof(GPUDrawable),
  };

  ContainerInput container = {
    .id = 0x01,
    .size = {225, 155},
    .offset = {0, 0},
    .anchor = ANCHOR_TOP_RIGHT,
    .layers = &layer,
    .layer_count = 1,
  };

  return create_container(&container, &context->render, &context->ui);
}

VkResult update_region_info_ui(ClientContext* context) { 
  char temp[20];
  VkResult result;
  EditorData* data = context->app_data;
  HexRegion* selected_region = context->hex.regions[data->selected_region];


  if(selected_region == NULL) {
    snprintf(temp,
        sizeof(temp),
        "");
    VK_RESULT(update_ui_string(temp, 0x01, 0, 0, &context->ui, &context->render));

    snprintf(
        temp,
        sizeof(temp),
        "");
    VK_RESULT(update_ui_string(temp, 0x01, 0, 1, &context->ui, &context->render));

    snprintf(
        temp,
        sizeof(temp),
        "");
    VK_RESULT(update_ui_string(temp, 0x01, 0, 2, &context->ui, &context->render));

    snprintf(
        temp,
        sizeof(temp),
        "");
    VK_RESULT(update_ui_string(temp, 0x01, 0, 3, &context->ui, &context->render));
  } else {
    snprintf(temp,
        sizeof(temp),
        "Region %4d", data->selected_region);
    VK_RESULT(update_ui_string(temp, 0x01, 0, 0, &context->ui, &context->render));

    snprintf(
        temp,
        sizeof(temp),
        "Q: %4d", selected_region->data.position.q);
    VK_RESULT(update_ui_string(temp, 0x01, 0, 1, &context->ui, &context->render));

    snprintf(
        temp,
        sizeof(temp),
        "R: %4d", selected_region->data.position.r);
    VK_RESULT(update_ui_string(temp, 0x01, 0, 2, &context->ui, &context->render));

    snprintf(
        temp,
        sizeof(temp),
        "Y: %4d", selected_region->data.y);
    VK_RESULT(update_ui_string(temp, 0x01, 0, 3, &context->ui, &context->render));
  }

  return VK_SUCCESS;
}

VkResult update_hex_info_ui(ClientContext* context) { 
  char temp[20];
  VkResult result;
  HexRegion* region = context->hex.regions[context->hex.data.clicked_region];


  if(region == NULL) {
    snprintf(
        temp,
        sizeof(temp),
        "");
    VK_RESULT(update_ui_string(temp, 0x02, 0, 0, &context->ui, &context->render));

    snprintf(
        temp,
        sizeof(temp),
        "");
    VK_RESULT(update_ui_string(temp, 0x02, 0, 1, &context->ui, &context->render));

    snprintf(
        temp,
        sizeof(temp),
        "");
    VK_RESULT(update_ui_string(temp, 0x02, 0, 2, &context->ui, &context->render));

    snprintf(
        temp,
        sizeof(temp),
        "");
    VK_RESULT(update_ui_string(temp, 0x02, 0, 3, &context->ui, &context->render));

    snprintf(
        temp,
        sizeof(temp),
        "");
    VK_RESULT(update_ui_string(temp, 0x02, 0, 4, &context->ui, &context->render));

    snprintf(
        temp,
        sizeof(temp),
        "");
    VK_RESULT(update_ui_string(temp, 0x02, 0, 5, &context->ui, &context->render));

    snprintf(
        temp,
        sizeof(temp),
        "");
    VK_RESULT(update_ui_string(temp, 0x02, 0, 6, &context->ui, &context->render));

    snprintf(
        temp,
        sizeof(temp),
        "");
    VK_RESULT(update_ui_string(temp, 0x02, 0, 7, &context->ui, &context->render));
  } else {
    GPUHex* hex = &region->data.hexes[context->hex.data.clicked_hex];
    snprintf(temp,
        sizeof(temp),
        "%d-%d",
        context->hex.data.clicked_region,
        context->hex.data.clicked_hex);
    VK_RESULT(update_ui_string(temp, 0x02, 0, 0, &context->ui, &context->render));

    HexCoord hex_world;
    hex_qr(context->hex.data.clicked_hex, &hex_world);
    
    HexCoord region_coord = {region->data.position.q, region->data.position.r};
    HexCoord region_world;
    region_qr(region_coord, &region_world);
    hex_add(region_world, &hex_world);

    snprintf(
        temp,
        sizeof(temp),
        "%d, %d, %d, %d",
        hex_world.q,
        hex_world.r,
        region->data.y,
        context->hex.data.clicked_vertex);
    VK_RESULT(update_ui_string(temp, 0x02, 0, 1, &context->ui, &context->render));

    snprintf(
        temp,
        sizeof(temp),
        "%02.02f %02.02f %02.02f",
        hex->height[0], hex->height[1], hex->height[2]);
    VK_RESULT(update_ui_string(temp, 0x02, 0, 2, &context->ui, &context->render));

    snprintf(
        temp,
        sizeof(temp),
        "%02.02f %02.02f %02.02f",
        hex->height[3], hex->height[4], hex->height[5]);
    VK_RESULT(update_ui_string(temp, 0x02, 0, 3, &context->ui, &context->render));

    snprintf(
        temp,
        sizeof(temp),
        "%08X",
        hex->color[0]);
    VK_RESULT(update_ui_string(temp, 0x02, 0, 4, &context->ui, &context->render));

    snprintf(
        temp,
        sizeof(temp),
        "%08X %08X",
        hex->color[1], hex->color[2]);
    VK_RESULT(update_ui_string(temp, 0x02, 0, 5, &context->ui, &context->render));

    snprintf(
        temp,
        sizeof(temp),
        "%08X %08X",
        hex->color[3], hex->color[4]);
    VK_RESULT(update_ui_string(temp, 0x02, 0, 6, &context->ui, &context->render));

    snprintf(
        temp,
        sizeof(temp),
        "%08X %08X",
        hex->color[5], hex->color[6]);
    VK_RESULT(update_ui_string(temp, 0x02, 0, 7, &context->ui, &context->render));
  }

  return VK_SUCCESS;
}

bool editor_key_callback(ClientContext* context, int key, int action, int mods) {
  uint32_t start, next_region;

  if(action != GLFW_PRESS) return false;

  EditorData* data = context->app_data;
  HexRegion* region = NULL;

  switch(key) {
  case GLFW_KEY_EQUAL:
    data->selected_region = add_hex_region(context);
    update_region_info_ui(context);
    update_hex_info_ui(context);
    break;

  case GLFW_KEY_MINUS:
    free_hex_region(data->selected_region, &context->hex, &context->render);
    start = data->selected_region;
    next_region = start;
    do {
      next_region -= 1;
      if(next_region > MAX_LOADED_REGIONS) next_region = MAX_LOADED_REGIONS;
      if(next_region != MAX_LOADED_REGIONS && context->hex.regions[next_region] != NULL) break;
    } while(next_region != start);
    if(next_region == start) {
      data->selected_region = MAX_LOADED_REGIONS;
    } else {
      data->selected_region = next_region;
    }
    update_region_info_ui(context);
    update_hex_info_ui(context);
    break;

  case GLFW_KEY_LEFT_BRACKET:
    start = data->selected_region;
    next_region = start;
    do {
      next_region -= 1;
      if(next_region > MAX_LOADED_REGIONS) next_region = MAX_LOADED_REGIONS;
      if(next_region != MAX_LOADED_REGIONS && context->hex.regions[next_region] != NULL) break;
    } while(next_region != start);
    data->selected_region = next_region;
    update_region_info_ui(context);
    break;

  case GLFW_KEY_RIGHT_BRACKET:
    start = data->selected_region;
    next_region = start;
    do {
      next_region += 1;
      if(next_region > MAX_LOADED_REGIONS) next_region = 0;
      if(next_region != MAX_LOADED_REGIONS && context->hex.regions[next_region] != NULL) break;
    } while(next_region != start);
    data->selected_region = next_region;
    update_region_info_ui(context);
    break;

  case GLFW_KEY_I:
    if(context->hex.regions[data->selected_region] != NULL) {
      context->hex.regions[data->selected_region]->data.position.q += 1;
      add_transfer(
          &context->hex.regions[data->selected_region]->data.position,
          context->hex.regions[data->selected_region]->region,
          offsetof(GPUHexRegion, position),
          sizeof(uint32_t),
          context->render.current_frame,
          &context->render);
    }
    update_region_info_ui(context);
    break;
  case GLFW_KEY_K:
    if(context->hex.regions[data->selected_region] != NULL) {
      context->hex.regions[data->selected_region]->data.position.q -= 1;
      add_transfer(
          &context->hex.regions[data->selected_region]->data.position.q,
          context->hex.regions[data->selected_region]->region,
          offsetof(GPUHexRegion, position) + offsetof(HexCoord, q),
          sizeof(uint32_t),
          context->render.current_frame,
          &context->render);
    }
    update_region_info_ui(context);
    break;
  case GLFW_KEY_J:
    if(context->hex.regions[data->selected_region] != NULL) {
      context->hex.regions[data->selected_region]->data.position.r += 1;
      add_transfer(
          &context->hex.regions[data->selected_region]->data.position.r,
          context->hex.regions[data->selected_region]->region,
          offsetof(GPUHexRegion, position) + offsetof(HexCoord, r),
          sizeof(uint32_t),
          context->render.current_frame,
          &context->render);
    }
    update_region_info_ui(context);
    break;
  case GLFW_KEY_L:
    if(context->hex.regions[data->selected_region] != NULL) {
      context->hex.regions[data->selected_region]->data.position.r -= 1;
      add_transfer(
          &context->hex.regions[data->selected_region]->data.position.r,
          context->hex.regions[data->selected_region]->region,
          offsetof(GPUHexRegion, position) + offsetof(HexCoord, r),
          sizeof(uint32_t),
          context->render.current_frame,
          &context->render);
    }
    update_region_info_ui(context);
    break;
  case GLFW_KEY_O:
    if(context->hex.regions[data->selected_region] != NULL) {
      context->hex.regions[data->selected_region]->data.y += 1;
      add_transfer(
          &context->hex.regions[data->selected_region]->data.y,
          context->hex.regions[data->selected_region]->region,
          offsetof(GPUHexRegion, y),
          sizeof(int32_t),
          context->render.current_frame,
          &context->render);
    }
    update_region_info_ui(context);
    break;
  case GLFW_KEY_U:
    if(context->hex.regions[data->selected_region] != NULL) {
      context->hex.regions[data->selected_region]->data.y -= 1;
      add_transfer(
          &context->hex.regions[data->selected_region]->data.y,
          context->hex.regions[data->selected_region]->region,
          offsetof(GPUHexRegion, y),
          sizeof(int32_t),
          context->render.current_frame,
          &context->render);
    }
    update_region_info_ui(context);
    break;

  case GLFW_KEY_Y:
    region = context->hex.regions[context->hex.data.clicked_region];
    if(region != NULL && context->hex.data.clicked_vertex != 0) {
      float height = region->data.hexes[context->hex.data.clicked_hex].height[context->hex.data.clicked_vertex-1] + 0.1; 

      if(~mods & GLFW_MOD_CONTROL) {
        uint32_t n_vertex[2], n_hex[2];
        HexCoord n_region[2];
        HexCoord world, region_center;

        region_qr(region->data.position, &region_center);
        hex_qr(context->hex.data.clicked_hex, &world);
        hex_add(region_center, &world);
        hex_vertex_neighbors(context->hex.data.clicked_vertex, world, n_vertex, n_region, n_hex);  

        uint32_t n0, n1;
        first_matching_region(n_region[0], region->data.y, &n0, &context->hex);
        first_matching_region(n_region[1], region->data.y, &n1, &context->hex);

        if(n0 != MAX_LOADED_REGIONS) set_vertex_height(height, n0, n_hex[0], n_vertex[0], context);
        if(n1 != MAX_LOADED_REGIONS) set_vertex_height(height, n1, n_hex[1], n_vertex[1], context);
      }
      set_vertex_height(height, context->hex.data.clicked_region, context->hex.data.clicked_hex, context->hex.data.clicked_vertex, context);
      update_hex_info_ui(context);
    }
    break;

  case GLFW_KEY_H:
    region = context->hex.regions[context->hex.data.clicked_region];
    if(region != NULL && context->hex.data.clicked_vertex != 0) {
      float height = region->data.hexes[context->hex.data.clicked_hex].height[context->hex.data.clicked_vertex-1] - 0.1; 

      if(~mods & GLFW_MOD_CONTROL) {
        uint32_t n_vertex[2], n_hex[2];
        HexCoord n_region[2];
        HexCoord world, region_center;

        region_qr(region->data.position, &region_center);
        hex_qr(context->hex.data.clicked_hex, &world);
        hex_add(region_center, &world);
        hex_vertex_neighbors(context->hex.data.clicked_vertex, world, n_vertex, n_region, n_hex);  

        uint32_t n0, n1;
        first_matching_region(n_region[0], region->data.y, &n0, &context->hex);
        first_matching_region(n_region[1], region->data.y, &n1, &context->hex);

        if(n0 != MAX_LOADED_REGIONS) set_vertex_height(height, n0, n_hex[0], n_vertex[0], context);
        if(n1 != MAX_LOADED_REGIONS) set_vertex_height(height, n1, n_hex[1], n_vertex[1], context);
      }
      set_vertex_height(height, context->hex.data.clicked_region, context->hex.data.clicked_hex, context->hex.data.clicked_vertex, context);
      update_hex_info_ui(context);
    }
    break;

  case GLFW_KEY_E:
    region = context->hex.regions[context->hex.data.clicked_region];
    if(region != NULL) {
      uint32_t n_vertex[2], n_hex[2];
      HexCoord n_region[2];
      HexCoord world, region_center;

      if(context->hex.data.clicked_vertex != 0) {
        region_qr(region->data.position, &region_center);
        hex_qr(context->hex.data.clicked_hex, &world);
        hex_add(region_center, &world);
        hex_vertex_neighbors(context->hex.data.clicked_vertex, world, n_vertex, n_region, n_hex);  

        uint32_t n0, n1;
        first_matching_region(n_region[0], region->data.y, &n0, &context->hex);
        first_matching_region(n_region[1], region->data.y, &n1, &context->hex);

        if(n0 != MAX_LOADED_REGIONS) set_vertex_color(data->current, n0, n_hex[0], n_vertex[0], context);
        if(n1 != MAX_LOADED_REGIONS) set_vertex_color(data->current, n1, n_hex[1], n_vertex[1], context);
      }
      set_vertex_color(data->current, context->hex.data.clicked_region, context->hex.data.clicked_hex, context->hex.data.clicked_vertex, context);
      update_hex_info_ui(context);
    }
    break;
  }

  return false;
}

void editor_frame(ClientContext* context) {
  EditorData* data = context->app_data;
  if(data->last_clicked_region != context->hex.data.clicked_region ||
     data->last_clicked_hex != context->hex.data.clicked_hex ||
     data->last_clicked_vertex != context->hex.data.clicked_vertex) {
    data->last_clicked_region = context->hex.data.clicked_region;
    data->last_clicked_hex = context->hex.data.clicked_hex;
    data->last_clicked_vertex = context->hex.data.clicked_vertex;
    update_hex_info_ui(context);
  }
}

void editor_startup(ClientContext* context) {
  region_info_ui(context);
  hex_info_ui(context);
  color_ui(context);
}

int main() {
  EditorData* data = malloc(sizeof(EditorData));
  memset(data, 0, sizeof(EditorData));
  data->selected_region = MAX_LOADED_REGIONS;
  data->last_clicked_region = MAX_LOADED_REGIONS;

  return run_app(data, editor_startup, editor_frame, NULL, editor_key_callback, NULL, NULL, NULL);
}
